# Copyright 2017 theloop Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from aio_pika.robust_channel import RobustChannel
from pika.adapters.blocking_connection import BlockingChannel


class MessageQueueInfoAsync:
    def __init__(self, channel: RobustChannel, route_key: str=None):
        self._channel = channel
        self._route_key = route_key

    async def queue_info(self, name: str=None):
        if name is None:
            name = self._route_key

        return await self._channel.declare_queue(name, passive=True)

    async def exchange_info(self, name: str=None):
        return await self._channel.declare_exchange(name, passive=True)


class MessageQueueInfoSync:
    def __init__(self, channel: BlockingChannel, route_key: str):
        self._channel = channel
        self._route_key = route_key

    def queue_info(self, name: str=None):
        if name is None:
            name = self._route_key

        return self._channel.queue_declare(queue=name, passive=True)

    def exchange_info(self, name: str=None):
        if name is None:
            name = self._route_key

        return self._channel.exchange_declare(exchange=name, passive=True)



# Copyright 2017 theloop Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# The original codes exist in aio_pika.patterns.rpc

import asyncio
import time
from typing import Callable

from aio_pika.exchange import ExchangeType
from aio_pika.channel import Channel
from aio_pika.message import Message, IncomingMessage
from aio_pika.patterns.base import Base


class Server(Base):
    DLX_NAME = 'rpc.dlx'

    def __init__(self, channel: Channel, queue_name):
        self.channel = channel

        self.func_names = {}
        self.routes = {}

        self.queue_name = queue_name
        self.queue = None

        self.dlx_exchange = None

    @asyncio.coroutine
    def initialize_exchange(self):
        self.dlx_exchange = yield from self.channel.declare_exchange(
            self.DLX_NAME,
            type=ExchangeType.HEADERS,
            auto_delete=True,
        )

    @asyncio.coroutine
    def initialize_queue(self, **kwargs):
        arguments = kwargs.pop('arguments', {}).update({
            'x-dead-letter-exchange': self.DLX_NAME,
        })

        kwargs['arguments'] = arguments

        self.queue = yield from self.channel.declare_queue(name=self.queue_name, **kwargs)

    def create_callback(self, func_name, func):
        if func_name in self.routes:
            raise RuntimeError(
                'Method name already used for %r' % self.routes[func_name]
            )

        self.func_names[func] = func_name
        self.routes[func_name] = func

    @asyncio.coroutine
    def consume(self):
        yield from self.queue.consume(self.on_callback)

    @asyncio.coroutine
    def on_callback(self, message: IncomingMessage):
        func_name = message.headers['FuncName']
        if func_name not in self.routes:
            return

        payload = self.deserialize(message.body)
        func = self.routes[func_name]

        try:
            result = yield from self._execute(func, payload)
            result = self.serialize(result)
            message_type = 'result'
        except Exception as e:
            result = self.serialize(e)
            message_type = 'error'

        result_message = Message(
            result,
            delivery_mode=message.delivery_mode,
            correlation_id=message.correlation_id,
            timestamp=time.time(),
            type=message_type,
        )

        yield from self.channel.default_exchange.publish(
            result_message,
            message.reply_to,
            mandatory=False
        )

        message.ack()

    @asyncio.coroutine
    def _execute(self, func, payload):
        return (yield from func(**payload))





# Copyright 2017 theloop Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from typing import Optional, TYPE_CHECKING

import aio_pika

from .message_queue_info import MessageQueueInfoAsync

if TYPE_CHECKING:
    from aio_pika.robust_connection import RobustChannel, RobustConnection


class MessageQueueConnection:
    def __init__(self, amqp_target, route_key, username=None, password=None):
        self._amqp_target = amqp_target
        self._route_key = route_key

        self._username = username or os.getenv("AMQP_USERNAME", "guest")
        self._password = password or os.getenv("AMQP_PASSWORD", "guest")

        self._connection: 'RobustConnection' = None
        self._channel: 'RobustChannel' = None

        self._async_info: MessageQueueInfoAsync = None

    async def connect(self, connection_attempts=None, retry_delay=None):
        kwargs = {}
        if connection_attempts is not None:
            kwargs['connection_attempts'] = connection_attempts
        if retry_delay is not None:
            kwargs['retry_delay'] = retry_delay

        self._connection: 'RobustConnection' = await aio_pika.connect_robust(
            host=self._amqp_target,
            login=self._username,
            password=self._password,
            **kwargs)

        self._connection.add_close_callback(self._callback_connection_close)
        self._connection.add_reconnect_callback(self._callback_connection_reconnect_callback)

        self._channel: 'RobustChannel' = await self._connection.channel()

        self._async_info = MessageQueueInfoAsync(self._channel, self._route_key)

    def async_info(self):
        return self._async_info

    def _callback_connection_close(self, sender, exc: Optional[BaseException], *args, **kwargs):
        pass

    def _callback_connection_reconnect_callback(self, sender, connection: 'RobustConnection', *args, **kwargs):
        pass



import unittest


class CallbackCollection:
    def __init__(self):
        self._cb: list = []
        self._sender = "sender"

    def add(self, cb):
        self._cb.append(cb)

    def __call__(self, *args, **kwargs):
        for cb in self._cb:
            cb(self._sender, *args, **kwargs)


class CallbackObj:
    def callback(self, sender, msg):
        print(sender, msg)


class TestCallback(unittest.TestCase):
    def test_func(self):
        collection = CallbackCollection()
        obj = CallbackObj()
        collection.add(obj.callback)
        collection("msg")



# Copyright 2017 theloop Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# The original codes exist in aio_pika.patterns.master

from typing import TYPE_CHECKING

import asyncio
from typing import Callable
from aio_pika.channel import Channel
from aio_pika.message import DeliveryMode
from aio_pika.patterns.base import Base

if TYPE_CHECKING:
    from aio_pika.message import IncomingMessage


class Server(Base):
    CONTENT_TYPE = 'application/python-pickle'
    DELIVERY_MODE = DeliveryMode.PERSISTENT

    def __init__(self, channel: Channel, queue_name):
        self.channel = channel
        self.queue_name = queue_name

        self.routes = {}
        self.queue = None

    @asyncio.coroutine
    def initialize_queue(self, **kwargs):
        self.queue = yield from self.channel.declare_queue(name=self.queue_name, **kwargs)

    def create_callback(self, func_name: str, func: Callable):
        self.routes[func_name] = func

    @asyncio.coroutine
    def consume(self):
        yield from self.queue.consume(self.on_callback)

    @asyncio.coroutine
    def on_callback(self, message: 'IncomingMessage'):
        func_name = message.headers['FuncName']
        func = self.routes.get(func_name)
        if func:
            with message.process(requeue=True, ignore_processed=True):
                data = self.deserialize(message.body)
                yield from self._execute(func, data)

    @classmethod
    @asyncio.coroutine
    def _execute(cls, func, kwargs):
        kwargs = kwargs or {}
        result = yield from func(**kwargs)
        return result



# Copyright 2017 theloop Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# The original codes exist in aio_pika.patterns.rpc

import time
from concurrent import futures

from aio_pika import ExchangeType, Message, DeliveryMode
from aio_pika.patterns.base import Base
from pika import BasicProperties


class ClientSync(Base):
    DLX_NAME = 'rpc.dlx'

    def __init__(self, channel, queue_name):
        self.channel = channel
        self.queue_name = queue_name
        self.result_queue_name = None

        self.futures = {}

        self.func_names = {}
        self.routes = {}

        self.dlx_exchange = None

    def initialize_exchange(self):
        self.dlx_exchange = self.channel.exchange_declare(
            exchange=self.DLX_NAME,
            exchange_type=ExchangeType.HEADERS.value,
            auto_delete=True,
        )

    def initialize_queue(self, **kwargs):
        arguments = kwargs.pop('arguments', {}).update({
            'x-dead-letter-exchange': self.DLX_NAME,
        })

        kwargs['arguments'] = arguments

        self.channel.queue_declare(queue=self.queue_name, **kwargs)
        self.result_queue_name = self.channel.queue_declare(queue='', exclusive=True, auto_delete=True).method.queue

        self.channel.queue_bind(
            queue=self.queue_name,
            exchange=self.DLX_NAME,
            arguments={
                "From": self.result_queue_name,
                'x-match': 'any',
            }
        )
        self.channel.basic_consume(on_message_callback=self._on_result_message, queue=self.result_queue_name, auto_ack=False)

    def _on_result_message(self, channel, method, properties, body):
        correlation_id = int(properties.correlation_id)
        try:
            future = self.futures[correlation_id]
        except KeyError:
            pass
        else:
            payload = self.deserialize(body)
            future.set_result(payload)

    def call(self, func_name, kwargs: dict=None, *, expiration: int=None,
             priority: int=128, delivery_mode: DeliveryMode=DeliveryMode.NOT_PERSISTENT):
        future = self._create_future()
        correlation_id = id(future)

        message = Message(
            body=self.serialize(kwargs or {}),
            type='call',
            timestamp=time.time(),
            expiration=expiration,
            priority=priority,
            correlation_id=correlation_id,
            delivery_mode=delivery_mode,
            reply_to=self.result_queue_name,
            headers={
                'From': self.result_queue_name,
                'FuncName': func_name
            },
        )

        properties = BasicProperties(
            message.properties.content_type,
            message.content_encoding,
            message.headers,
            message.delivery_mode,
            message.priority,
            message.correlation_id,
            message.reply_to,
            message.expiration,
            message.message_id,
            None,
            message.type,
            message.user_id,
            message.app_id,
        )

    # noinspection PyProtectedMember
        self.channel.basic_publish(
            '', self.queue_name, message.body, properties, mandatory=True
        )

        while not future.done():
            self.channel.connection.process_data_events()

        if future.exception():
            return future.exception()
        else:
            return future.result()

    def _create_future(self) -> futures.Future:
        future = futures.Future()
        future_id = id(future)
        self.futures[future_id] = future
        future.add_done_callback(lambda f: self.futures.pop(future_id, None))
        return future



# Copyright 2017 theloop Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from typing import TypeVar, Generic

from . import MessageQueueConnection, MessageQueueType, MESSAGE_QUEUE_TYPE_KEY, TASK_ATTR_DICT
from .patterns import worker, rpc

T = TypeVar('T')


class MessageQueueService(MessageQueueConnection, Generic[T]):
    TaskType: type = object

    loop = asyncio.get_event_loop() or asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    def __init__(self, amqp_target, route_key, username=None, password=None, **task_kwargs):
        super().__init__(amqp_target, route_key, username, password)

        if self.TaskType is object and type(self) is not MessageQueueService:
            raise RuntimeError("MessageQueueTasks is not specified.")

        self._worker_server: worker.Server = None
        self._rpc_server: rpc.Server = None

        self._task = self.__class__.TaskType(**task_kwargs)

    async def connect(self, connection_attempts=None, retry_delay=None, **kwargs):
        await super().connect(connection_attempts, retry_delay)

        self._worker_server = worker.Server(self._channel, self._route_key)
        self._rpc_server = rpc.Server(self._channel, self._route_key)

        queue = await self._channel.declare_queue(self._route_key, auto_delete=True)
        await queue.consume(self._consume, **kwargs)

        await self._serve_tasks()

    async def _serve_tasks(self):
        for attribute_name in dir(self._task):
            try:
                attribute = getattr(self._task, attribute_name)
                task_attr: dict = getattr(attribute, TASK_ATTR_DICT)
            except AttributeError:
                pass
            else:
                func_name = f"{type(self._task).__name__}.{attribute_name}"

                message_queue_type = task_attr[MESSAGE_QUEUE_TYPE_KEY]
                if message_queue_type == MessageQueueType.Worker:
                    self._worker_server.create_callback(func_name, attribute)
                elif message_queue_type == MessageQueueType.RPC:
                    self._rpc_server.create_callback(func_name, attribute)
                else:
                    raise RuntimeError(f"MessageQueueType invalid. {func_name}, {message_queue_type}")

    async def _consume(self, message):
        await self._worker_server.on_callback(message)
        await self._rpc_server.on_callback(message)

    def serve(self, **kwargs):
        self.loop.create_task(self.connect(**kwargs))

    @classmethod
    def serve_all(cls):
        cls.loop.run_forever()



# Copyright 2017 theloop Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# The original codes exist in aio_pika.patterns.rpc


import asyncio
import logging
import time
from typing import TYPE_CHECKING

from aio_pika.channel import Channel
from aio_pika.exceptions import DeliveryError
from aio_pika.exchange import ExchangeType
from aio_pika.message import Message, DeliveryMode
from aio_pika.patterns.base import Base

if TYPE_CHECKING:
    from aiormq.types import DeliveredMessage
    from aio_pika.message import IncomingMessage


class ClientAsync(Base):
    DLX_NAME = 'rpc.dlx'

    def __init__(self, channel: Channel, queue_name):
        self.channel = channel
        self.queue_name = queue_name
        self.queue = None
        self.result_queue = None

        self.async_futures = {}
        self.concurrent_futures = {}

        self.func_names = {}
        self.routes = {}

        self.dlx_exchange = None

    @asyncio.coroutine
    def initialize_exchange(self):
        self.dlx_exchange = yield from self.channel.declare_exchange(
            self.DLX_NAME,
            type=ExchangeType.HEADERS,
            auto_delete=True,
        )

    @asyncio.coroutine
    def initialize_queue(self, **kwargs):
        arguments = kwargs.pop('arguments', {}).update({
            'x-dead-letter-exchange': self.DLX_NAME,
        })

        kwargs['arguments'] = arguments

        self.queue = yield from self.channel.declare_queue(name=self.queue_name, **kwargs)

        self.result_queue = yield from self.channel.declare_queue(None, exclusive=True, auto_delete=True)
        yield from self.result_queue.bind(
            self.dlx_exchange, "",
            arguments={
                "From": self.result_queue.name,
                'x-match': 'any',
            }
        )

        yield from self.result_queue.consume(
            self._on_result_message, no_ack=True
        )

        self.channel.add_on_return_callback(self._on_message_returned)

    def _on_message_returned(self, sender, message: 'DeliveredMessage', *args, **kwargs):
        correlation_id = int(message.correlation_id) if message.correlation_id else None

        future = self.async_futures.pop(correlation_id, None) or self.concurrent_futures.pop(correlation_id, None)
        if future and future.done():
            logging.warning(f"Unknown message was returned: ({sender}) {message}")
        else:
            future.set_exception(DeliveryError(message, None))

    @asyncio.coroutine
    def _on_result_message(self, message: 'IncomingMessage'):
        correlation_id = int(message.correlation_id) if message.correlation_id else None
        try:
            future = self.async_futures[correlation_id]  # type: asyncio.Future
        except KeyError:
            pass
        else:
            payload = self.deserialize(message.body)

            if message.type == 'result':
                future.set_result(payload)
            elif message.type == 'error':
                future.set_exception(payload)
            elif message.type == 'call':
                future.set_exception(asyncio.TimeoutError("Message timed-out", message))
            else:
                future.set_exception(RuntimeError("Unknown message type %r" % message.type))

    @asyncio.coroutine
    def call(self, func_name, kwargs: dict=None, *, expiration: int=None,
             priority: int=128, delivery_mode: DeliveryMode=DeliveryMode.NOT_PERSISTENT):
        future = self._create_future()
        message = Message(
            body=self.serialize(kwargs or {}),
            type='call',
            timestamp=time.time(),
            expiration=expiration,
            priority=priority,
            correlation_id=id(future),
            delivery_mode=delivery_mode,
            reply_to=self.result_queue.name,
            headers={
                'From': self.result_queue.name,
                'FuncName': func_name
            },
        )

        yield from self.channel.default_exchange.publish(
            message, routing_key=self.queue_name, mandatory=True
        )

        return (yield from future)

    def _create_future(self) -> asyncio.Future:
        future = self.channel.loop.create_future()
        future_id = id(future)
        self.async_futures[future_id] = future
        future.add_done_callback(lambda f: self.async_futures.pop(future_id, None))
        return future



import asyncio
import unittest
import random
import string

from earlgrey import MessageQueueStub, MessageQueueService, MessageQueueType, message_queue_task


class TestBasic(unittest.TestCase):
    def test_order(self):
        order = list(reversed(range(8)))

        def _assert_order(num):
            curr_order = order.pop()
            print(f"curr_order : {curr_order}")
            self.assertEqual(num, curr_order)

        class Task:
            @message_queue_task(type_=MessageQueueType.RPC)
            async def long_time_rpc(self, x, y):
                _assert_order(1)
                await asyncio.sleep(0.5)
                _assert_order(2)
                return x + y

            @message_queue_task(type_=MessageQueueType.Worker)
            async def long_time_work(self):
                await asyncio.sleep(0.5)
                _assert_order(6)

        class Stub(MessageQueueStub[Task]):
            TaskType = Task

        class Service(MessageQueueService[Task]):
            TaskType = Task

        async def _run():
            route_key = 'any same string between processes'

            client = Stub('localhost', route_key)
            server = Service('localhost', route_key)

            await client.connect()
            await server.connect()

            _assert_order(0)
            result = await client.async_task().long_time_rpc(10, 20)
            _assert_order(3)
            self.assertEqual(result, 30)

            _assert_order(4)
            await client.async_task().long_time_work()
            _assert_order(5)

            await asyncio.sleep(1.5)
            _assert_order(7)

            self.assertEqual(0, len(order))

        loop = asyncio.get_event_loop()
        loop.run_until_complete(_run())

    def test_mq_info(self):
        class Task:
            @message_queue_task(type_=MessageQueueType.Worker)
            async def work_async(self):
                pass

            @message_queue_task(type_=MessageQueueType.Worker)
            def work_sync(self):
                pass

        class Stub(MessageQueueStub[Task]):
            TaskType = Task

        class Service(MessageQueueService[Task]):
            TaskType = Task

        async def _run():
            route_key_length = random.randint(5, 10)
            route_key = ''.join(random.choice(string.ascii_letters) for _ in range(route_key_length))

            client = Stub('localhost', route_key)
            await client.connect()

            await client.async_task().work_async()
            client.sync_task().work_sync()
            await asyncio.sleep(1)

            info = await client.async_info().queue_info()
            self.assertEqual(info.declaration_result.message_count, 2)

            info = client.sync_info().queue_info()
            self.assertEqual(info.method.message_count, 2)

            server = Service('localhost', route_key)
            await server.connect()
            await asyncio.sleep(1)

            info = await client.async_info().queue_info()
            self.assertEqual(info.declaration_result.message_count, 0)

            info = client.sync_info().queue_info()
            self.assertEqual(info.method.message_count, 0)

        loop = asyncio.get_event_loop()
        loop.run_until_complete(_run())





# Copyright 2017 theloop Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# The original codes exist in aio_pika.patterns.master

import asyncio
import logging
from typing import TYPE_CHECKING

from aio_pika.channel import Channel
from aio_pika.message import Message, DeliveryMode
from aio_pika.patterns.base import Base

if TYPE_CHECKING:
    from aiormq.types import DeliveredMessage


class ClientAsync(Base):
    CONTENT_TYPE = 'application/python-pickle'
    DELIVERY_MODE = DeliveryMode.PERSISTENT

    def __init__(self, channel: Channel, queue_name):
        self.channel = channel
        self.queue_name = queue_name
        self.queue = None

        self.channel.add_on_return_callback(self._on_message_returned)

    @asyncio.coroutine
    def initialize_queue(self, **kwargs):
        self.queue = yield from self.channel.declare_queue(name=self.queue_name, **kwargs)

    @asyncio.coroutine
    def call(self, func_name: str, kwargs=None, priority=128):
        message = Message(
            body=self.serialize(kwargs or {}),
            content_type=self.CONTENT_TYPE,
            delivery_mode=self.DELIVERY_MODE,
            priority=priority,
            headers={
                'FuncName': func_name
            }
        )

        yield from self.channel.default_exchange.publish(
            message, self.queue_name, mandatory=True
        )

    @classmethod
    def _on_message_returned(cls, sender, message: 'DeliveredMessage', *args, **kwargs):
        logging.warning(
            f"Message returned. Probably destination queue does not exists: ({sender}) {message}"
        )



import asyncio
import unittest
import multiprocessing
import random
import string
import subprocess
import traceback
import os

from earlgrey import MessageQueueStub, MessageQueueService, MessageQueueType, message_queue_task


def available_credential(username, password):
    try:
        subprocess.check_output(["rabbitmqctl", "authenticate_user", username, password])
        result = subprocess.check_output(["rabbitmqctl", "list_user_permissions", username])
        return b"/	.*	.*	.*" in result
    except Exception as e:
        traceback.print_exc()
        return False


class TestMultiprocess(unittest.TestCase):
    def test_basic(self):
        server = multiprocessing.Process(target=self._run_server)
        server.daemon = True
        server.start()

        client = multiprocessing.Process(target=self._run_client)
        client.daemon = True
        client.start()

        server.join()
        client.join()

        self.assertEqual(server.exitcode, 0)
        self.assertEqual(client.exitcode, 0)

    def test_credential_failure(self):
        random_length = random.randrange(5, 10)
        random_string = "".join(random.choice(string.ascii_letters) for _ in range(random_length))

        print(f"username : {random_string}")
        print(f"password : {random_string}")

        server = multiprocessing.Process(target=self._run_server, args=(random_string, random_string))
        server.daemon = True
        server.start()

        client = multiprocessing.Process(target=self._run_client, args=(random_string, random_string))
        client.daemon = True
        client.start()

        server.join()
        client.join()

        self.assertEqual(server.exitcode, 1)
        self.assertEqual(client.exitcode, 1)

    test_username = "test"
    test_password = "test"

    @unittest.skipUnless(available_credential(test_username, test_password),
                         f"{test_username} / {test_password} invalid")
    def test_credential_success(self):
        print(f"username : {self.test_username}")
        print(f"password : {self.test_password}")

        server = multiprocessing.Process(target=self._run_server, args=(self.test_username, self.test_password))
        server.daemon = True
        server.start()

        client = multiprocessing.Process(target=self._run_client, args=(self.test_username, self.test_password))
        client.daemon = True
        client.start()

        server.join()
        client.join()

        self.assertEqual(server.exitcode, 0)
        self.assertEqual(client.exitcode, 0)

    @unittest.skipUnless(available_credential(test_username, test_password),
                         f"{test_username} / {test_password} invalid")
    def test_credential_success_environment_variable(self):
        print(f"username : {self.test_username}")
        print(f"password : {self.test_password}")

        os.environ["AMQP_USERNAME"] = self.test_username
        os.environ["AMQP_PASSWORD"] = self.test_password

        server = multiprocessing.Process(target=self._run_server)
        server.daemon = True
        server.start()

        client = multiprocessing.Process(target=self._run_client)
        client.daemon = True
        client.start()

        server.join()
        client.join()

        self.assertEqual(server.exitcode, 0)
        self.assertEqual(client.exitcode, 0)

    def _run_server(self, username=None, password=None):
        async def _run():
            try:
                message_queue_service = Service('localhost', route_key, username, password)
                await message_queue_service.connect()
            except:
                loop.stop()
                raise

        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        task = loop.create_task(_run())
        loop.run_forever()

        if task.exception():
            raise task.exception()

        print("run_server finished.")

    def _run_client(self, username=None, password=None):
        async def _run():
            try:
                message_queue_stub = Stub('localhost', route_key, username, password)
                await message_queue_stub.connect()

                result = await message_queue_stub.async_task().sum(10, 20)
                self.assertEqual(result, 30)

                result = message_queue_stub.sync_task().multiply(10, 20)
                self.assertEqual(result, 200)

                message_queue_stub.sync_task().ping(123)

                await message_queue_stub.async_task().stop()
            except:
                loop.stop()
                raise

        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        loop.run_until_complete(_run())
        print("run_client finished.")


class Task:
    @message_queue_task
    async def sum(self, x, y):
        return x + y

    @message_queue_task
    def multiply(self, x, y):
        return x * y

    @message_queue_task(type_=MessageQueueType.Worker)
    def ping(self, value):
        print(f'value : {value}')
        assert value == 123

    @message_queue_task(type_=MessageQueueType.Worker)
    async def stop(self):
        print('stop')
        asyncio.get_event_loop().stop()


class Stub(MessageQueueStub[Task]):
    TaskType = Task


class Service(MessageQueueService[Task]):
    TaskType = Task


route_key = 'something same you want'


# Copyright 2017 theloop Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import functools
import logging
import traceback

from enum import IntEnum

TASK_ATTR_DICT = "task_attr"
MESSAGE_QUEUE_TYPE_KEY = "type"
MESSAGE_QUEUE_PRIORITY_KEY = "priority"


class MessageQueueType(IntEnum):
    Worker = 0,
    RPC = 1,


class MessageQueueException(Exception):
    pass


def message_queue_task(func=None, *, type_=MessageQueueType.RPC, priority=128):
    if func is None:
        return functools.partial(message_queue_task, type_=type_, priority=priority)

    @functools.wraps(func)
    async def _wrapper(*args, **kwargs):
        try:
            return await asyncio.coroutine(func)(*args, **kwargs)
        except Exception as e:
            logging.error(e)
            traceback.print_exc()
            return MessageQueueException(str(e))

    task_attr = {
        MESSAGE_QUEUE_TYPE_KEY: type_,
        MESSAGE_QUEUE_PRIORITY_KEY: priority
    }
    setattr(_wrapper, TASK_ATTR_DICT, task_attr)
    return _wrapper



import sys
from importlib import import_module
from pathlib import Path

from setuptools import setup, find_packages

sys.path.insert(0, str(Path.cwd() / 'earlgrey'))
try:
    module = import_module('version')
    version = getattr(module, '__version__')
finally:
    sys.path = sys.path[1:]

extras_requires = {
    'test': ['pytest~=5.4.2', 'mock~=4.0.1']
}

setup_options = {
    'name': 'earlgrey',
    'description': 'Python AMQP RPC library',
    'long_description': open('README.md').read(),
    'long_description_content_type': 'text/markdown',
    'url': 'https://github.com/icon-project/earlgrey',
    'version': version,
    'author': 'ICON foundation',
    'packages': find_packages(),
    'license': "Apache License 2.0",
    'install_requires': list(open('requirements.txt')),
    'extras_require': extras_requires,
    'classifiers': [
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Natural Language :: English',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ]
}

setup(**setup_options)



# Copyright 2017 theloop Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# The original codes exist in aio_pika.patterns.master

from aio_pika.message import DeliveryMode, Message
from aio_pika.patterns.base import Base
from pika.adapters.blocking_connection import BlockingChannel
from pika.spec import BasicProperties


class ClientSync(Base):
    CONTENT_TYPE = 'application/python-pickle'
    DELIVERY_MODE = DeliveryMode.PERSISTENT

    def __init__(self, channel: BlockingChannel, queue_name):
        self.channel = channel
        self.queue_name = queue_name

    def initialize_queue(self, **kwargs):
        self.channel.queue_declare(queue=self.queue_name, **kwargs)

    def call(self, func_name: str, kwargs=None, priority=128):
        message = Message(
            body=self.serialize(kwargs or {}),
            content_type=self.CONTENT_TYPE,
            delivery_mode=self.DELIVERY_MODE,
            priority=priority,
            headers={
                'FuncName': func_name
            }
        )

        properties = BasicProperties(
            message.properties.content_type,
            message.content_encoding,
            message.headers,
            message.delivery_mode,
            message.priority,
            message.correlation_id,
            message.reply_to,
            message.expiration,
            message.message_id,
            None,
            message.type,
            message.user_id,
            message.app_id,
        )
        # noinspection PyProtectedMember
        self.channel.basic_publish(
            '',
            self.queue_name,
            message.body,
            properties
        )



# Copyright 2017 theloop Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import inspect
import logging
import threading
from typing import TypeVar, Generic

import pika

from . import (MessageQueueConnection, MessageQueueInfoSync, MessageQueueType, MessageQueueException, worker, rpc,
               MESSAGE_QUEUE_TYPE_KEY, MESSAGE_QUEUE_PRIORITY_KEY, TASK_ATTR_DICT)

T = TypeVar('T')


class MessageQueueStub(MessageQueueConnection, Generic[T]):
    TaskType: type = object

    def __init__(self, amqp_target, route_key, username=None, password=None):
        super().__init__(amqp_target, route_key, username, password)

        if self.TaskType is object and type(self) is not MessageQueueStub:
            raise RuntimeError("MessageQueueTasks is not specified.")

        self._worker_client_async: worker.ClientAsync = None
        self._rpc_client_async: rpc.ClientAsync = None

        self._async_task = object.__new__(self.__class__.TaskType)  # not calling __init__
        self._thread_local = _Local()

    async def connect(self, connection_attempts=None, retry_delay=None):
        await super().connect(connection_attempts, retry_delay)
        await self._connect_async()
        self._register_tasks_async()

    async def _connect_async(self):
        self._worker_client_async = worker.ClientAsync(self._channel, self._route_key)
        await self._worker_client_async.initialize_queue(auto_delete=True)

        self._rpc_client_async = rpc.ClientAsync(self._channel, self._route_key)
        await self._rpc_client_async.initialize_exchange()
        await self._rpc_client_async.initialize_queue(auto_delete=True)

    def _connect_sync(self):
        credential_params = pika.PlainCredentials(self._username, self._password)
        connection_params = pika.ConnectionParameters(
            host=f'{self._amqp_target}',
            heartbeat=0,
            credentials=credential_params)
        
        connection = pika.BlockingConnection(connection_params)
        channel = connection.channel()

        self._thread_local.worker_client_sync = worker.ClientSync(channel, self._route_key)
        self._thread_local.worker_client_sync.initialize_queue(auto_delete=True)

        self._thread_local.rpc_client_sync: rpc.ClientSync = rpc.ClientSync(channel, self._route_key)
        self._thread_local.rpc_client_sync.initialize_exchange()
        self._thread_local.rpc_client_sync.initialize_queue(auto_delete=True)

        self._thread_local.sync_task = object.__new__(self.__class__.TaskType)  # not calling __init__
        self._thread_local.sync_info = MessageQueueInfoSync(channel, self._route_key)

    def _register_tasks_async(self):
        for attribute_name in dir(self._async_task):
            try:
                attribute = getattr(self._async_task, attribute_name)
                task_attr: dict = getattr(attribute, TASK_ATTR_DICT)
            except AttributeError:
                pass
            else:
                func_name = f"{type(self._async_task).__name__}.{attribute_name}"

                message_queue_type = task_attr[MESSAGE_QUEUE_TYPE_KEY]
                message_queue_priority = task_attr[MESSAGE_QUEUE_PRIORITY_KEY]
                if message_queue_type == MessageQueueType.Worker:
                    binding_async_method = self._call_async_worker
                elif message_queue_type == MessageQueueType.RPC:
                    binding_async_method = self._call_async_rpc
                else:
                    raise RuntimeError(f"MessageQueueType invalid. {func_name}, {message_queue_type}")

                stub = functools.partial(binding_async_method, func_name, attribute, message_queue_priority)
                setattr(self._async_task, attribute_name, stub)

    def _register_tasks_sync(self):
        for attribute_name in dir(self._thread_local.sync_task):
            try:
                attribute = getattr(self._thread_local.sync_task, attribute_name)
                task_attr: dict = getattr(attribute, TASK_ATTR_DICT)
            except AttributeError:
                pass
            else:
                func_name = f"{type(self._thread_local.sync_task).__name__}.{attribute_name}"

                message_queue_type = task_attr[MESSAGE_QUEUE_TYPE_KEY]
                message_queue_priority = task_attr[MESSAGE_QUEUE_PRIORITY_KEY]
                if message_queue_type == MessageQueueType.Worker:
                    binding_sync_method = self._call_sync_worker
                elif message_queue_type == MessageQueueType.RPC:
                    binding_sync_method = self._call_sync_rpc
                else:
                    raise RuntimeError(f"MessageQueueType invalid. {func_name}, {message_queue_type}")

                stub = functools.partial(binding_sync_method, func_name, attribute, message_queue_priority)
                setattr(self._thread_local.sync_task, attribute_name, stub)

    async def _call_async_worker(self, func_name, func, priority, *args, **kwargs):
        params = inspect.signature(func).bind(*args, **kwargs)
        params.apply_defaults()
        await self._worker_client_async.call(func_name, kwargs=params.arguments, priority=priority)

    async def _call_async_rpc(self, func_name, func, priority, *args, **kwargs):
        params = inspect.signature(func).bind(*args, **kwargs)
        params.apply_defaults()
        result = await self._rpc_client_async.call(func_name, kwargs=params.arguments, priority=priority)
        if isinstance(result, MessageQueueException):
            logging.error(result)
            raise result
        return result

    def _call_sync_worker(self, func_name, func, priority, *args, **kwargs):
        params = inspect.signature(func).bind(*args, **kwargs)
        params.apply_defaults()
        self._thread_local.worker_client_sync.call(func_name, kwargs=params.arguments, priority=priority)

    def _call_sync_rpc(self, func_name, func, priority, *args, **kwargs):
        params = inspect.signature(func).bind(*args, **kwargs)
        params.apply_defaults()
        result = self._thread_local.rpc_client_sync.call(func_name, kwargs=params.arguments, priority=priority)
        if isinstance(result, MessageQueueException):
            logging.error(result)
            raise result
        return result

    def async_task(self) -> T:
        return self._async_task

    def sync_task(self) -> T:
        if self._thread_local.sync_task is None:
            self._connect_sync()
            self._register_tasks_sync()

        return self._thread_local.sync_task

    def sync_info(self):
        if self._thread_local.sync_info is None:
            self._connect_sync()
            self._register_tasks_sync()

        return self._thread_local.sync_info


class _Local(threading.local):
    worker_client_sync: worker.ClientSync = None
    rpc_client_sync: rpc.ClientSync = None
    sync_task = None
    sync_info: MessageQueueInfoSync = None
